@@ -1,14 +1,18 @@
 import logging
 import os
+from functools import lru_cache
 from pathlib import Path
 from typing import Dict, Optional
 
 import black
 import toml
 from pylsp import hookimpl
+from pylsp._utils import get_eol_chars
+from pylsp.config.config import Config
 
 logger = logging.getLogger(__name__)
 
+
 GLOBAL_CONFIG: Optional[Path] = None
 try:
     if os.name == "nt":
@@ -25,19 +29,34 @@ else:
 
 
 @hookimpl(tryfirst=True)
-def pylsp_format_document(document):
-    return format_document(document)
+def pylsp_format_document(config, document):
+    return format_document(config, document)
 
 
 @hookimpl(tryfirst=True)
-def pylsp_format_range(document, range):
+def pylsp_format_range(config, document, range):
     range["start"]["character"] = 0
     range["end"]["line"] += 1
     range["end"]["character"] = 0
-    return format_document(document, range)
+    return format_document(config, document, range)
 
 
-def format_document(document, range=None):
+@hookimpl
+def pylsp_settings():
+    """Configuration options that can be set on the client."""
+    return {
+        "plugins": {
+            "black": {
+                "enabled": True,
+                "line_length": 88,
+                "preview": False,
+                "cache_config": False,
+            }
+        }
+    }
+
+
+def format_document(client_config, document, range=None):
     if range:
         start = range["start"]["line"]
         end = range["end"]["line"]
@@ -49,7 +68,7 @@ def format_document(document, range=None):
             "end": {"line": len(document.lines), "character": 0},
         }
 
-    config = load_config(document.path)
+    config = load_config(document.path, client_config)
 
     try:
         formatted_text = format_text(text=text, config=config)
@@ -68,8 +87,25 @@ def format_text(*, text, config):
         string_normalization=not config["skip_string_normalization"],
     )
     try:
-        # will raise black.NothingChanged, we want to bubble that exception up
-        return black.format_file_contents(text, fast=config["fast"], mode=mode)
+        # Black's format_file_contents only works reliably when eols are '\n'. It gives
+        # an error for '\r' and produces wrong formatting for '\r\n'. So we replace
+        # those eols by '\n' before formatting and restore them afterwards.
+        replace_eols = False
+        eol_chars = get_eol_chars(text)
+        if eol_chars is not None and eol_chars != "\n":
+            replace_eols = True
+            text = text.replace(eol_chars, "\n")
+
+        # Will raise black.NothingChanged, we want to bubble that exception up
+        formatted_text = black.format_file_contents(
+            text, fast=config["fast"], mode=mode
+        )
+
+        # Restore eols if necessary.
+        if replace_eols:
+            formatted_text = formatted_text.replace("\n", eol_chars)
+
+        return formatted_text
     except (
         # raised when the file has syntax errors
         ValueError,
@@ -84,13 +120,17 @@ def format_text(*, text, config):
         raise black.NothingChanged from e
 
 
-def load_config(filename: str) -> Dict:
+@lru_cache(100)
+def _load_config(filename: str, client_config: Config) -> Dict:
+    settings = client_config.plugin_settings("black")
+
     defaults = {
-        "line_length": 88,
+        "line_length": settings.get("line_length", 88),
         "fast": False,
         "pyi": filename.endswith(".pyi"),
         "skip_string_normalization": False,
         "target_version": set(),
+        "preview": settings.get("preview", False),
     }
 
     root = black.find_project_root((filename,))
@@ -139,6 +179,7 @@ def load_config(filename: str) -> Dict:
             black.TargetVersion.PY37,
             black.TargetVersion.PY38,
             black.TargetVersion.PY39,
+            black.TargetVersion.PY310,
         }
     else:
         target_version = set()
@@ -148,3 +189,13 @@ def load_config(filename: str) -> Dict:
     logger.info("Using config from %s: %r", pyproject_filename, config)
 
     return config
+
+
+def load_config(filename: str, client_config: Config) -> Dict:
+    settings = client_config.plugin_settings("black")
+
+    # Use the original, not cached function to load settings if requested
+    if not settings.get("cache_config", False):
+        return _load_config.__wrapped__(filename, client_config)
+
+    return _load_config(filename, client_config)
